"""
pyrad.prod.process_product
==========================

Functions for obtaining Pyrad products from the datasets

.. autosummary::
    :toctree: generated/

    generate_occurrence_products
    generate_icon_coord_products
    generate_icon_to_radar_products
    generate_sun_hits_products
    generate_qvp_products
    generate_ml_products
    generate_vpr_products
    generate_centroids_products

"""

import matplotlib.pyplot as plt
from copy import deepcopy
from warnings import warn
import os

import numpy as np

import pyart

from .process_vol_products import generate_vol_products

from ..io.io_aux import get_fieldname_pyart
from ..io.io_aux import get_save_dir, make_filename

from ..io.read_data_sun import read_sun_retrieval
from ..io.read_data_other import read_ml_ts

from ..io.write_data import write_sun_hits, write_sun_retrieval
from ..io.write_data import write_excess_gates, write_ts_ml, write_histogram
from ..io.write_data import write_timeseries_point, write_centroids
from ..io.write_data import write_rhi_profile, write_vpr_theo_params
from ..io.write_data import write_vpr_info

from ..graph.plots import plot_sun_hits, plot_histogram2, plot_scatter
from ..graph.plots import plot_centroids
from ..graph.plots_timeseries import plot_sun_retrieval_ts, plot_ml_ts
from ..graph.plots_vol import plot_fixed_rng_sun
from ..graph.plots_vol import plot_along_coord

from ..util.radar_utils import create_sun_hits_field, compute_histogram
from ..util.radar_utils import create_sun_retrieval_field

import matplotlib as mpl

mpl.use("Agg")

# Increase a bit font size
mpl.rcParams.update({"font.size": 16})
mpl.rcParams.update({"font.family": "sans-serif"})


def generate_occurrence_products(dataset, prdcfg):
    """
    generates occurrence products. Accepted product types:
        'WRITE_EXCESS_GATES': Write the data that identifies radar gates
            with clutter that has a frequency of occurrence above a certain
            threshold.
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                quant_min: float
                    Minimum frequency of occurrence in percentage to keep the
                    gate as valid. Default 95.
        All the products of the 'VOL' dataset group

    Parameters
    ----------
    dataset : tuple
        radar object and metadata dictionary

    prdcfg : dictionary of dictionaries
        product configuration dictionary of dictionaries

    Returns
    -------
    filename : str
        the name of the file created. None otherwise

    """
    instant = False
    if "instant" in prdcfg:
        instant = prdcfg["instant"]

    if not instant and not dataset["occu_final"]:
        return None

    if prdcfg["type"] == "WRITE_EXCESS_GATES":
        if not dataset["occu_final"]:
            return None

        radar = dataset["radar_out"]
        if (
            ("frequency_of_occurrence" not in radar.fields)
            or ("occurrence" not in radar.fields)
            or ("number_of_samples" not in radar.fields)
        ):
            warn("Unable to create quantile excess gates file. " "Missing data")
            return None

        dssavedir = prdcfg["dsname"]
        if "dssavename" in prdcfg:
            dssavedir = prdcfg["dssavename"]

        quant_min = 95.0
        if "quant_min" in prdcfg:
            quant_min = prdcfg["quant_min"]

        # get index of gates exceeding quantile
        freq_occu = radar.fields["frequency_of_occurrence"]["data"]
        ind_ray, ind_rng = np.where(freq_occu > quant_min)
        if ind_ray.size == 0:
            warn("No data exceeds the frequency of occurrence " + str(quant_min) + " %")
            return None

        excess_dict = {
            "starttime": dataset["starttime"],
            "endtime": dataset["endtime"],
            "quant_min": quant_min,
            "ray_ind": ind_ray,
            "rng_ind": ind_rng,
            "ele": radar.elevation["data"][ind_ray],
            "azi": radar.azimuth["data"][ind_ray],
            "rng": radar.range["data"][ind_rng],
            "nsamples": (radar.fields["number_of_samples"]["data"][ind_ray, ind_rng]),
            "occurrence": (radar.fields["occurrence"]["data"][ind_ray, ind_rng]),
            "freq_occu": freq_occu[ind_ray, ind_rng],
        }
        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=dataset["endtime"],
        )

        fname = make_filename(
            "excess_gates",
            prdcfg["dstype"],
            prdcfg["prdname"],
            ["csv"],
            prdcfginfo="quant" + "{:.1f}".format(quant_min),
            timeinfo=dataset["endtime"],
        )

        fname = savedir + fname[0]

        fname = write_excess_gates(excess_dict, fname)

        if fname is not None:
            print("saved excess gates file: {}".format(fname))

        return fname

    field_name = get_fieldname_pyart(prdcfg["voltype"])
    if (field_name == "frequency_of_occurrence") and (not dataset["occu_final"]):
        return None
    if dataset["occu_final"]:
        prdcfg["timeinfo"] = dataset["endtime"]

    return generate_vol_products(dataset, prdcfg)


def generate_icon_coord_products(dataset, prdcfg):
    """
    generates icon coordinates products. Accepted product types:
        'SAVEVOL': Save an object containing the index of the icon model grid
            that corresponds to each radar gate in a C/F radial file.
            User defined parameters:
                file_type: str
                    The type of file used to save the data. Can be 'nc' or
                    'h5'. Default 'nc'
                physical: Bool
                    If True the data will be saved in physical units (floats).
                    Otherwise it will be quantized and saved as binary
                compression: str
                    For ODIM file formats, the type of compression. Can be any
                    of the allowed compression types for hdf5 files. Default
                    gzip
                compression_opts: any
                    The compression options allowed by the hdf5. Depends on
                    the type of compression. Default 6 (The gzip compression
                    level).

    Parameters
    ----------
    dataset : tuple
        radar object containing the icon coordinates

    prdcfg : dictionary of dictionaries
        product configuration dictionary of dictionaries

    Returns
    -------
    filename : str
        the name of the file created. None otherwise

    """
    if prdcfg["type"] == "SAVEVOL":
        radar_obj = dataset["radar_out"]
        ind_rad = dataset["ind_rad"]

        field_name = get_fieldname_pyart(prdcfg["voltype"])
        if field_name not in radar_obj.fields:
            warn(
                " Field type "
                + field_name
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        file_type = prdcfg.get("file_type", "nc")
        prdcfg.get("physical", True)
        compression = prdcfg.get("compression", "gzip")
        compression_opts = prdcfg.get("compression_opts", 6)

        new_dataset = deepcopy(radar_obj)
        new_dataset.fields = dict()
        new_dataset.add_field(field_name, radar_obj.fields[field_name])

        savedir = prdcfg["iconpath"][ind_rad] + "rad2icon/"
        fname = "rad2icon_" + prdcfg["voltype"] + "_" + prdcfg["procname"] + ".nc"

        if file_type == "nc":
            pyart.io.cfradial.write_cfradial(savedir + fname, new_dataset)
        elif file_type == "h5":
            pyart.aux_io.write_odim_h5(
                savedir + fname,
                new_dataset,
                compression=compression,
                compression_opts=compression_opts,
            )
        else:
            warn("Data could not be saved. " + "Unknown saving file type " + file_type)
            return None

        print("saved file: {}".format(savedir + fname))

        return fname

    warn(" Unsupported product type: " + prdcfg["type"])
    return None


def generate_icon_to_radar_products(dataset, prdcfg):
    """
    generates icon data in radar coordinates products. Accepted product
    types:
        'SAVEVOL': Save an object containing the icon data in radar
            coordinatesin a C/F radial or ODIM file.
                User defined parameters:
                file_type: str
                    The type of file used to save the data. Can be 'nc' or
                    'h5'. Default 'nc'
                physical: Bool
                    If True the data will be saved in physical units (floats).
                    Otherwise it will be quantized and saved as binary
                compression: str
                    For ODIM file formats, the type of compression. Can be any
                    of the allowed compression types for hdf5 files. Default
                    gzip
                compression_opts: any
                    The compression options allowed by the hdf5. Depends on
                    the type of compression. Default 6 (The gzip compression
                    level).
        All the products of the 'VOL' dataset group

    Parameters
    ----------
    dataset : tuple
        radar object containing the icon coordinates

    prdcfg : dictionary of dictionaries
        product configuration dictionary of dictionaries

    Returns
    -------
    filename : str
        the name of the file created. None otherwise

    """
    time_index = prdcfg.get("icon_time_index", 0)
    if time_index > len(dataset) - 1:
        warn(
            "icon time index larger than available. Skipping product " + prdcfg["type"]
        )
        return None

    radar_dataset = dataset[time_index]
    if prdcfg["type"] == "SAVEVOL":
        radar_obj = radar_dataset["radar_out"]
        ind_rad = radar_dataset["ind_rad"]

        field_name = get_fieldname_pyart(prdcfg["voltype"])
        if field_name not in radar_obj.fields:
            warn(
                " Field type "
                + field_name
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        file_type = prdcfg.get("file_type", "nc")
        prdcfg.get("physical", True)
        compression = prdcfg.get("compression", "gzip")
        compression_opts = prdcfg.get("compression_opts", 6)

        new_dataset = deepcopy(radar_obj)
        new_dataset.fields = dict()
        new_dataset.add_field(field_name, radar_obj.fields[field_name])

        savedir = (
            prdcfg["iconpath"][ind_rad]
            + prdcfg["voltype"]
            + "/radar/"
            + prdcfg["timeinfo"].strftime("%Y-%m-%d")
            + "/"
            + prdcfg["procname"]
            + "/"
        )
        fname = (
            prdcfg["voltype"]
            + "_RUN"
            + prdcfg["timeinfo"].strftime("%Y%m%d%H%M%S")
            + "_"
            + radar_dataset["dticon"].strftime("%Y%m%d%H%M%S")
            + ".nc"
        )

        if not os.path.isdir(savedir):
            os.makedirs(savedir)

        if file_type == "nc":
            pyart.io.cfradial.write_cfradial(savedir + fname, new_dataset)
        elif file_type == "h5":
            pyart.aux_io.write_odim_h5(
                savedir + fname,
                new_dataset,
                compression=compression,
                compression_opts=compression_opts,
            )
        else:
            warn("Data could not be saved. " + "Unknown saving file type " + file_type)
            return None

        print("saved file: {}".format(savedir + fname))

        return fname

    return generate_vol_products(radar_dataset, prdcfg)


def generate_sun_hits_products(dataset, prdcfg):
    """
    generates sun hits products. Accepted product types:
        All the products of the 'VOL' dataset group
        'PLOT_SUN_HITS': Plots in a sun-radar azimuth difference-sun-radar
            elevation difference grid the values of all sun hits obtained
            during the processing period
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
        'PLOT_SUN_RETRIEVAL': Plots in a sun-radar azimuth difference-sun-
            radar elevation difference grid the retrieved sun pattern
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
        'PLOT_SUN_RETRIEVAL_TS': Plots time series of the retrieved sun
            pattern parameters
            User defined parameters:
                voltype: str
                    name of the pyrad variable to use, it must be available in the dataset
                dpi: int
                    The pixel density of the plot. Default 72
                add_date_in_fname: Bool
                    If true the year is added in the plot file name
        'PLOT_SUNSCAN': Plots a constant range radar azimuth-elevation of the
            sunscan field data
        'WRITE_SUN_HITS': Writes the information concerning possible sun hits
            in a csv file. The written data has the following columns
            "time",
            "ray",
            "NPrng",
            "rad_el",
            "rad_az",
            "sun_el",
            "sun_az",
            "dBm_sun_hit",
            "std(dBm_sun_hit)",
            "NPh",
            "NPhval",
            "dBmv_sun_hit",
            "std(dBmv_sun_hit)",
            "NPv",
            "NPvval",
            "ZDR_sun_hit",
            "std(ZDR_sun_hit)",
            "NPzdr",
            "NPzdrval"
        'WRITE_SUN_RETRIEVAL': Writes the retrieved sun pattern parameters in
            a csv file.
            User defined parameters:
                add_date_in_fname: Bool
                    If true the year is added in the csv file name
        'WRITE_SUNSCAN': Writes the sunscan parameters in a csv file. The written data has the following columns
            "sun_az":             [deg] Azimuth sun position,
            "sun_el":             [deg] Elevation sun position,
            "noise_pwr":          [dBm] Noise power",
            "sun_maxpwr_noise":   [dBm] sun maximal power sample (including noise),
            "sun_maxpwr_nonoise": [dBm] sun maximal power sample without noise,
            "sun_maxpwr_fit":     [dBm] sun maximal fitted power (without noise),
            "sun_maxpwr_toa":     [dBm] sun maximal power at top of atmosphere,
            "az_offset":          [deg] Azimuth shift of fitted maxima to sun azimuth,
            "el_offset":          [deg] Elevation shift of fitted maxima to sun elevation,
            "az_phi3db":          [deg] Half-power beam width in azimuth,
            "el_phi3db":          [deg] Half-power beam width in elevation,
            "fit_stddev":         [dBm] Standard deviation (fit to samples),
            "num_samples":        [#]   Number of samples used for the sun power fitting

    Parameters
    ----------
    dataset : tuple
        radar object and sun hits dictionary

    prdcfg : dictionary of dictionaries
        product configuration dictionary of dictionaries

    Returns
    -------
    filename : str
        the name of the file created. None otherwise

    """

    dssavedir = prdcfg["dsname"]
    if "dssavename" in prdcfg:
        dssavedir = prdcfg["dssavename"]

    prdcfg["timeinfo"] = dataset["timeinfo"]

    if prdcfg["type"] == "WRITE_SUN_HITS":
        if "sun_hits" not in dataset:
            return None

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=dataset["timeinfo"],
        )

        fname = make_filename(
            "info",
            prdcfg["dstype"],
            "detected",
            ["csv"],
            timeinfo=dataset["timeinfo"],
            timeformat="%Y%m%d",
        )[0]

        fname = savedir + fname

        write_sun_hits(dataset["sun_hits"], fname)

        print("saved sun hits file: {}".format(fname))

        return fname[0]

    if prdcfg["type"] == "PLOT_SUN_HITS":
        if "sun_hits_final" not in dataset:
            return None

        field_name = get_fieldname_pyart(prdcfg["voltype"])

        if prdcfg["voltype"] not in dataset["sun_hits_final"]:
            warn(
                " Field type "
                + prdcfg["voltype"]
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=dataset["timeinfo"],
        )

        fname_list = make_filename(
            "detected",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            timeinfo=dataset["timeinfo"],
            timeformat="%Y%m%d",
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        field = create_sun_hits_field(
            dataset["sun_hits_final"]["rad_el"],
            dataset["sun_hits_final"]["rad_az"],
            dataset["sun_hits_final"]["sun_el"],
            dataset["sun_hits_final"]["sun_az"],
            dataset["sun_hits_final"][prdcfg["voltype"]],
            prdcfg["sunhitsImageConfig"],
        )

        if field is None:
            warn(
                "Unable to create field "
                + prdcfg["voltype"]
                + " Skipping product "
                + prdcfg["type"]
            )
            return None

        plot_sun_hits(field, field_name, fname_list, prdcfg)

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if prdcfg["type"] == "WRITE_SUN_RETRIEVAL":
        if "sun_retrieval" not in dataset:
            return None

        timeinfo = None
        timeformat = None
        if prdcfg.get("add_date_in_fname", False):
            timeinfo = dataset["timeinfo"]
            timeformat = "%Y"

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=None,
        )

        fname = make_filename(
            "info",
            prdcfg["dstype"],
            "retrieval",
            ["csv"],
            timeinfo=timeinfo,
            timeformat=timeformat,
            runinfo=prdcfg["runinfo"],
        )[0]

        fname = savedir + fname

        write_sun_retrieval(dataset["sun_retrieval"], fname)

        print("saved sun retrieval file: {}".format(fname))

        return fname

    if prdcfg["type"] == "PLOT_SUN_RETRIEVAL":
        if "sun_retrieval" not in dataset:
            return None

        field_name = get_fieldname_pyart(prdcfg["voltype"])
        par = None
        if field_name == "sun_est_power_h":
            par = "par_h"
        elif field_name == "sun_est_power_v":
            par = "par_v"
        elif field_name == "sun_est_differential_reflectivity":
            par = "par_zdr"

        if par not in dataset["sun_retrieval"]:
            warn(
                " Field type "
                + prdcfg["voltype"]
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=dataset["timeinfo"],
        )

        fname_list = make_filename(
            "retrieval",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            timeinfo=dataset["timeinfo"],
            timeformat="%Y%m%d",
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        if dataset["sun_retrieval"][par] is None:
            warn(" Invalid retrieval parameters. Skipping product " + prdcfg["type"])
            return None

        field = create_sun_retrieval_field(
            dataset["sun_retrieval"][par],
            field_name,
            prdcfg["sunhitsImageConfig"],
            lant=dataset["sun_retrieval"]["lant"],
        )

        if field is not None:
            plot_sun_hits(field, field_name, fname_list, prdcfg)

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if prdcfg["type"] == "PLOT_SUN_RETRIEVAL_TS":
        if "sun_retrieval" not in dataset:
            return None

        dpi = prdcfg.get("dpi", 72)
        timeinfo = None
        timeformat = None
        if prdcfg.get("add_date_in_fname", False):
            timeinfo = dataset["timeinfo"]
            timeformat = "%Y"

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdid"],
            timeinfo=None,
        )

        fname = make_filename(
            "info",
            prdcfg["dstype"],
            "retrieval",
            ["csv"],
            timeinfo=timeinfo,
            timeformat=timeformat,
            runinfo=prdcfg["runinfo"],
        )

        fname = savedir + fname[0]

        sun_retrieval = read_sun_retrieval(fname)

        if sun_retrieval[0] is None:
            warn("Unable to read sun retrieval file " + fname)
            return None

        if len(sun_retrieval[0]) < 2:
            warn(
                "Unable to plot sun retrieval time series. " + "Not enough data points."
            )
            return None

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=None,
        )

        fname_list = make_filename(
            "retrieval_ts",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            timeinfo=timeinfo,
            timeformat=timeformat,
            runinfo=prdcfg["runinfo"],
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        titl = (
            prdcfg["runinfo"]
            + " Sun Retrieval "
            + sun_retrieval[1][0].strftime("%Y%m%d")
            + "-"
            + sun_retrieval[1][-1].strftime("%Y%m%d")
        )
        figfname = plot_sun_retrieval_ts(
            sun_retrieval, prdcfg["voltype"], fname_list, titl=titl, dpi=dpi
        )

        if figfname is None:
            return None

        print("----- save to " + " ".join(fname_list))
        return fname_list

    if prdcfg["type"] == "WRITE_SUNSCAN":
        if "sun_retrieval" not in dataset:
            return None

        text = [
            "SunScan info",
            "sun_az:             [deg] Azimuth sun position ",
            "sun_el:             [deg] Elevation sun position",
            "noise_pwr:          [dBm] Noise power",
            "sun_maxpwr_noise:   [dBm]" " sun maximal power sample (including noise)",
            "sun_maxpwr_nonoise: [dBm]" " sun maximal power sample without noise",
            "sun_maxpwr_fit:     [dBm]" " sun maximal fitted power (without noise)",
            "sun_maxpwr_toa:     [dBm]" " sun maximal power at top of atmosphere",
            "az_offset:          [deg]"
            " Azimuth shift of fitted maxima to sun azimuth",
            "el_offset:          [deg]"
            " Elevation shift of fitted maxima to sun elevation",
            "az_phi3db:          [deg]" " Half-power beam width in azimuth",
            "el_phi3db:          [deg]" " Half-power beam width in elevation",
            "fit_stddev:         [dBm]" " Standard deviation (fit to samples)",
            "num_samples:        [#]"
            " Number of samples used for the sun power fitting",
        ]

        sunRdata = dataset["sun_retrieval"]

        if dataset["field_name"] == "noisedBm_hh":
            data = {
                "dstype": prdcfg["dstype"],
                "unit": "dBm",
                "time": sunRdata["sunscan_time"],
                "label": [
                    "sun_az",
                    "sun_el",
                    "noise_pwr",
                    "sun_maxpwr_noise",
                    "sun_maxpwr_nonoise",
                    "sun_maxpwr_fit",
                    "sun_maxpwr_toa",
                    "az_offset",
                    "el_offset",
                    "az_phi3db",
                    "el_phi3db",
                    "fit_stddev",
                    "num_samples",
                ],
                "value": [
                    sunRdata["sunpos_az"],
                    sunRdata["sunpos_el"],
                    sunRdata["noise_pwr"],
                    sunRdata["sun_maxpwr_noise"],
                    sunRdata["sun_maxpwr_nonoise"],
                    sunRdata["dBm_sun_est"],
                    sunRdata["dBm_sun_est_toa"],
                    sunRdata["az_bias_h"],
                    sunRdata["el_bias_h"],
                    sunRdata["az_width_h"],
                    sunRdata["el_width_h"],
                    sunRdata["std(dBm_sun_est)"],
                    sunRdata["nhits_h"],
                ],
            }
        elif dataset["field_name"] == "noisedBm_vv":
            data = {
                "dstype": prdcfg["dstype"],
                "unit": "dBm",
                "time": sunRdata["sunscan_time"],
                "label": [
                    "sun_az",
                    "sun_el",
                    "noise_pwr",
                    "sun_maxpwr_noise",
                    "sun_maxpwr_nonoise",
                    "sun_maxpwr_fit",
                    "sun_maxpwr_toa",
                    "az_offset",
                    "el_offset",
                    "az_phi3db",
                    "el_phi3db",
                    "fit_stddev",
                    "num_samples",
                ],
                "value": [
                    sunRdata["sunpos_az"],
                    sunRdata["sunpos_el"],
                    sunRdata["noise_pwr"],
                    sunRdata["sun_maxpwr_noise"],
                    sunRdata["sun_maxpwr_nonoise"],
                    sunRdata["dBmv_sun_est"],
                    sunRdata["dBmv_sun_est_toa"],
                    sunRdata["az_bias_v"],
                    sunRdata["el_bias_v"],
                    sunRdata["az_width_v"],
                    sunRdata["el_width_v"],
                    sunRdata["std(dBmv_sun_est)"],
                    sunRdata["nhits_v"],
                ],
            }
        else:
            warn("ERROR: No valid datatype for WRITE_SUNSCAN product.")

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            prdcfg["timeinfo"],
        )

        fname1 = make_filename(
            "ts",
            prdcfg["dstype"],
            dataset["field_name"],
            ["csv"],
            timeinfo=prdcfg["timeinfo"],
            timeformat="%Y%m%d",
            runinfo=prdcfg["runinfo"],
        )[0]

        fname1 = savedir + fname1
        write_timeseries_point(fname1, data, prdcfg["dstype"], text)

        print("saved sunscan file: {}".format(fname1))

        return fname1

    if prdcfg["type"] == "PLOT_SUNSCAN":
        radar = dataset["radar_out"]
        sun_hits = dataset["sun_hits"]

        field_name = dataset["field_name"]
        if field_name not in radar.fields:
            warn(
                " Field type "
                + field_name
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        # user defined parameters
        prdcfg.get("azi_res", None)
        prdcfg.get("ele_res", None)
        vmin = prdcfg.get("vmin", None)
        vmax = prdcfg.get("vmax", None)
        angtol = prdcfg.get("ang_tol", 0.5)

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            prdcfg["timeinfo"],
        )

        fname_list = make_filename(
            "constr",
            prdcfg["dstype"],
            prdcfg["dsname"],
            prdcfg["imgformat"],
            prdcfginfo="rng" + "{:.1f}".format(dataset["radar_out"].range["data"][0]),
            timeinfo=prdcfg["timeinfo"],
            runinfo=prdcfg["runinfo"],
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        plot_fixed_rng_sun(
            radar,
            field_name,
            sun_hits,
            prdcfg,
            fname_list,
            azi_res=None,
            ele_res=None,
            ang_tol=angtol,
            vmin=vmin,
            vmax=vmax,
        )

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if "radar_out" in dataset:
        return generate_vol_products(dataset, prdcfg)

    return None


def generate_qvp_products(dataset, prdcfg):
    """
    Generates quasi vertical profile-like products, from azimuthal averaging of polarimetric radar data.  Accepted product types:
        All the products of the 'VOL' dataset group

    Parameters
    ----------
    dataset : dict
        dictionary containing the radar object and a keyword stating the
        status of the processing

    prdcfg : dictionary of dictionaries
        product configuration dictionary of dictionaries

    Returns
    -------
    filename : str
        the name of the file created. None otherwise

    """
    qvp_type = "final"
    if "qvp_type" in prdcfg:
        qvp_type = prdcfg["qvp_type"]

    if qvp_type == "final" and dataset["radar_type"] != "final":
        return None

    prdcfg["timeinfo"] = dataset["start_time"]
    return generate_vol_products(dataset, prdcfg)


def generate_ml_products(dataset, prdcfg):
    """
    Generates melting layer products. Accepted product types:
        'ML_TS': Plots and writes a time series of the melting layer, i.e.
            the evolution of the average and standard deviation of the melting
            layer top and thickness and the the number of rays used in the
            retrieval.
            User defined parameters:
                dpi: int
                    The pixel density of the plot. Default 72
        'SAVE_ML': Saves an object containing the best estimate of the melting
            layer retrieved in a C/F radial file
        'SAVE_RETRIEVED_ML': Saves an object containing the estimate of the
            melting layer at each elevation angle obtained by the MF retrieval
            in a C/F radial file
        All the products of the 'VOL' dataset group

    Parameters
    ----------
    dataset : dict
        dictionary containing the radar object and a keyword stating the
        status of the processing

    prdcfg : dictionary of dictionaries
        product configuration dictionary of dictionaries

    Returns
    -------
    filename : str
        the name of the file created. None otherwise

    """

    dssavedir = prdcfg["dsname"]
    if "dssavename" in prdcfg:
        dssavedir = prdcfg["dssavename"]

    if prdcfg["type"] == "ML_TS":
        dpi = prdcfg.get("dpi", 72)

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        csvfname = make_filename(
            "ts",
            prdcfg["dstype"],
            "ml",
            ["csv"],
            timeinfo=prdcfg["timeinfo"],
            timeformat="%Y%m%d",
        )[0]

        csvfname = savedir + csvfname

        ml_bottom = dataset["ml_obj"].fields["melting_layer_height"]["data"][:, 0]
        ml_top = dataset["ml_obj"].fields["melting_layer_height"]["data"][:, 1]

        ml_top_avg = np.ma.asarray(np.ma.mean(ml_top))
        ml_top_std = np.ma.asarray(np.ma.std(ml_top))
        thick = ml_top - ml_bottom
        thick_avg = np.ma.asarray(np.ma.mean(thick))
        thick_std = np.ma.asarray(np.ma.std(thick))
        nrays_valid = thick.compressed().size
        nrays_total = thick.size

        write_ts_ml(
            prdcfg["timeinfo"],
            ml_top_avg,
            ml_top_std,
            thick_avg,
            thick_std,
            nrays_valid,
            nrays_total,
            csvfname,
        )

        print("saved CSV file: {}".format(csvfname))

        (
            dt_ml_arr,
            ml_top_avg_arr,
            ml_top_std_arr,
            thick_avg_arr,
            thick_std_arr,
            nrays_valid_arr,
            nrays_total_arr,
        ) = read_ml_ts(csvfname)

        if dt_ml_arr is None:
            warn("Unable to plot time series. No valid data")
            return None

        figfname_list = make_filename(
            "ts",
            prdcfg["dstype"],
            "ml",
            prdcfg["imgformat"],
            timeinfo=dt_ml_arr[0],
            timeformat="%Y%m%d",
        )

        for i, figfname in enumerate(figfname_list):
            figfname_list[i] = savedir + figfname

        titl = dt_ml_arr[0].strftime("%Y-%m-%d") + " melting layer time series"

        plot_ml_ts(
            dt_ml_arr,
            ml_top_avg_arr,
            ml_top_std_arr,
            thick_avg_arr,
            thick_std_arr,
            nrays_valid_arr,
            nrays_total_arr,
            figfname_list,
            labelx="Time UTC",
            titl=titl,
            dpi=dpi,
        )
        print("----- save to " + " ".join(figfname_list))

        return figfname_list

    if prdcfg["type"] == "SAVE_ML":
        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        fname = make_filename(
            "saveml",
            prdcfg["dstype"],
            "ml_h",
            ["nc"],
            timeinfo=prdcfg["timeinfo"],
            runinfo=prdcfg["runinfo"],
        )[0]

        fname = savedir + fname
        pyart.io.cfradial.write_cfradial(fname, dataset["ml_obj"])
        print("saved file: {}".format(fname))

        return fname

    if prdcfg["type"] == "SAVE_RETRIEVED_ML":
        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        fname = make_filename(
            "saveml_retrieved",
            prdcfg["dstype"],
            "ml_h",
            ["nc"],
            timeinfo=prdcfg["timeinfo"],
            runinfo=prdcfg["runinfo"],
        )[0]

        fname = savedir + fname
        pyart.io.cfradial.write_cfradial(fname, dataset["ml_retrieved"])
        print("saved file: {}".format(fname))

        return fname

    return generate_vol_products(dataset, prdcfg)


def generate_vpr_products(dataset, prdcfg):
    """
    Generates melting layer products. Accepted product types:
        'PLOT_VPR_THEO': Plots and writes the retrieved theoretical VPR
            User defined parameters:
                dpi: int
                    The pixel density of the plot. Default 72
        'WRITE_VPR_THEO_PARAMS': Writes the parameters of the theoretical VPR
            in a text file
        'WRITE_VPR_INFO': Writes a dictionary containing relevant parameters
        of the VPR profile retrieval into a text file
            User defined parameters:
                    voltype: name of the pyrad variable to use, it must be available

        All the products of the 'VOL' dataset group

    Parameters
    ----------
    dataset : dict
        dictionary containing the radar object and a keyword stating the
        status of the processing

    prdcfg : dictionary of dictionaries
        product configuration dictionary of dictionaries

    Returns
    -------
    filename : str
        the name of the file created. None otherwise

    """

    dssavedir = prdcfg["dsname"]
    if "dssavename" in prdcfg:
        dssavedir = prdcfg["dssavename"]
    radar_type = prdcfg.get("radar_type", "radar_out")
    if radar_type == "radar_out":
        dataset_aux = {"radar_out": dataset["radar_out"]}
    else:
        dataset_aux = {"radar_out": dataset["radar_rhi"]}

    if prdcfg["type"] == "PLOT_VPR_THEO":
        dpi = prdcfg.get("dpi", 72)

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        csvfname = make_filename(
            "theo",
            prdcfg["dstype"],
            "VPR",
            ["csv"],
            timeinfo=prdcfg["timeinfo"],
            timeformat="%Y%m%d%H%M%S",
        )[0]

        csvfname = savedir + csvfname

        write_rhi_profile(
            dataset["vpr_theo_dict"]["altitude"],
            [dataset["vpr_theo_dict"]["value"]],
            None,
            ["Znorm"],
            csvfname,
            datatype="normalized reflectivity (-)",
            timeinfo=prdcfg["timeinfo"],
        )
        print("saved CSV file: {}".format(csvfname))

        figfname_list = make_filename(
            "theo",
            prdcfg["dstype"],
            "VPR",
            prdcfg["imgformat"],
            timeinfo=prdcfg["timeinfo"],
            timeformat="%Y%m%d%H%M%S",
        )

        for i, figfname in enumerate(figfname_list):
            figfname_list[i] = savedir + figfname

        titl = "{}\nRetrieved VPR profile".format(
            prdcfg["timeinfo"].strftime("%Y-%m-%dT%H:%M:%S")
        )
        plot_along_coord(
            [dataset["vpr_theo_dict"]["altitude"]],
            [dataset["vpr_theo_dict"]["value"]],
            figfname_list,
            labelx="Altitude (m MSL)",
            labely="normalized reflectivity (-)",
            labels=None,
            title=titl,
            colors=None,
            linestyles=None,
            ymin=None,
            ymax=None,
            dpi=dpi,
            data_on_y=False,
            plot_legend=False,
        )
        print("----- save to " + " ".join(figfname_list))

        return figfname_list

    if prdcfg["type"] == "WRITE_VPR_THEO_PARAMS":
        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        csvfname = make_filename(
            "theo",
            prdcfg["dstype"],
            "VPR",
            ["csv"],
            timeinfo=prdcfg["timeinfo"],
            timeformat="%Y%m%d%H%M%S",
        )[0]

        csvfname = savedir + csvfname

        write_vpr_theo_params(dataset["vpr_theo_dict"], csvfname)

        print("saved CSV file: {}".format(csvfname))

        return csvfname

    if prdcfg["type"] == "WRITE_VPR_INFO":
        if "vpr_info" not in dataset:
            warn("No VPR info available")
            return None

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=prdcfg["timeinfo"],
        )

        csvfname = make_filename(
            "info",
            prdcfg["dstype"],
            "VPR",
            ["csv"],
            timeinfo=prdcfg["timeinfo"],
            timeformat="%Y%m%d%H%M%S",
        )[0]

        csvfname = savedir + csvfname

        write_vpr_info(dataset["vpr_info"], csvfname)

        print("saved CSV file: {}".format(csvfname))

        return csvfname

    return generate_vol_products(dataset_aux, prdcfg)


def generate_centroids_products(dataset, prdcfg):
    """
    Generates centroids products. Accepted product types:
        'HISTOGRAM': Plots the histogram of one of the variables used for
            centroids computation.
            User defined parameters:
                voltype : str
                    The name of the variable to plot. Can be dBZ, ZDR, KDP,
                    RhoHV, H_ISO0 and its standardized form (e.g. dBZ_std)
                write_data : Bool
                    If true writes the histogram in a .csv file. Default True
                step : float
                    bin size. Default 0.1
        'HISTOGRAM2D': Plots the 2D- histogram of two of the variables used
            for centroids computation.
            User defined parameters:
                voltype_x, voltype_y : str
                    The name of the variables to plot. Can be dBZ, ZDR, KDP,
                    RhoHV, H_ISO0 and its standardized form (e.g. dBZ_std)
                step_x, step_y : float
                    bin size. Default 0.1
        'HISTOGRAM_LABELED': Plots the histogram of one of the variables used
            for centroids computation. Only plots labeled data.
            User defined parameters:
                voltype : str
                    The name of the variable to plot. Can be dBZ, ZDR, KDP,
                    RhoHV, H_ISO0 and its standardized form (e.g. dBZ_std)
                write_data : Bool
                    If true writes the histogram in a .csv file. Default True
                step : float
                    bin size. Default 0.1
        'HISTOGRAM_CENTROIDS': Plots the histogram of one of the variables
            used for centroids computation corresponding to a particular
            hydrometeor type, the intermediate centroids and the final
            centroid
            User defined parameters:
                voltype : str
                    The name of the variable to plot. Can be dBZ, ZDR, KDP,
                    RhoHV, H_ISO0 and its standardized form (e.g. dBZ_std)
                hydro_type : str
                    The name of the hydrometeor type.
                write_data : Bool
                    If true writes the histogram in a .csv file. Default True
                step : float
                    bin size. Default 0.1
        'HISTOGRAM2D_CENTROIDS': Plots the 2D- histogram of two of the
            variables used for centroids computatio ncorresponding to a
            particular hydrometeor type, the intermediate centroids and the
            final centroid
            User defined parameters:
                voltype_x, voltype_y : str
                    The name of the variables to plot. Can be dBZ, ZDR, KDP,
                    RhoHV, H_ISO0 and its standardized form (e.g. dBZ_std)
                hydro_type : str
                    The name of the hydrometeor type.
                step_x, step_y : float
                    bin size. Default 0.1
        'WRITE_CENTROIDS': Writes the final centroids in a .csv file.
        'SAVE_DATA': Saves the data used to compute the centroids in an .npz
            file
        'SAVE_LABELED_DATA': Saves the labeled data, the intermediate
            centroids and the final centroids in an .npz file


    Parameters
    ----------
    dataset : dict
        dictionary containing the radar object and a keyword stating the
        status of the processing

    prdcfg : dictionary of dictionaries
        product configuration dictionary of dictionaries

    Returns
    -------
    filename : str
        the name of the file created. None otherwise

    """
    dssavedir = prdcfg["dsname"]
    if "dssavename" in prdcfg:
        dssavedir = prdcfg["dssavename"]

    if prdcfg["type"] == "HISTOGRAM":
        data_dict = dataset["data_dict"]

        # check if we have to plot the data
        hist_type = prdcfg.get("hist_type", "cumulative")
        if hist_type == "cumulative" and not data_dict["final"]:
            return None
        if hist_type == "instant" and data_dict["final"]:
            return None

        voltype = prdcfg["voltype"]
        if voltype not in data_dict.keys():
            warn(
                " Field type "
                + voltype
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        write_data = prdcfg.get("write_data", 1)
        step = prdcfg.get("step", 0.1)

        timeformat = "%Y%m%d"
        timeinfo = data_dict["timeinfo"][0]
        titl = timeinfo.strftime("%Y-%m-%d") + "\n" + voltype
        data_vals = data_dict[voltype]
        if hist_type == "instant":
            timeformat = "%Y%m%d%H%M%S"
            timeinfo = data_dict["timeinfo"][-1]
            titl = timeinfo.strftime("%Y-%m-%d %H:%M:%S") + "\n" + voltype
            nvols = data_dict["npoints"].size
            ind_start = np.sum(data_dict["npoints"][0 : nvols - 2])
            data_vals = data_vals[ind_start:]

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=timeinfo,
        )

        fname_list = make_filename(
            "histogram",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            timeinfo=timeinfo,
            timeformat=timeformat,
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        if "_std" in voltype:
            bin_edges = np.arange(-1.0 - step / 2.0, 1.0 + step / 2 + step, step)
            values = data_vals
        else:
            field_name = get_fieldname_pyart(voltype)
            bin_edges, values = compute_histogram(data_vals, field_name, step=step)
        bin_centers = bin_edges[1:] - step / 2
        hist, bin_edges = np.histogram(values, bins=bin_edges)
        plot_histogram2(
            bin_centers,
            hist,
            fname_list,
            labelx=voltype,
            labely="Number of Samples",
            titl=titl,
        )

        print("----- save to " + " ".join(fname_list))

        if write_data:
            fname = (
                savedir
                + make_filename(
                    "histogram",
                    prdcfg["dstype"],
                    prdcfg["voltype"],
                    ["csv"],
                    timeinfo=timeinfo,
                    timeformat=timeformat,
                )[0]
            )

            write_histogram(bin_edges, hist, fname, step=step)
            print("----- save to {}".format(fname))

            return fname

        return fname_list

    if prdcfg["type"] == "HISTOGRAM2D":
        data_dict = dataset["data_dict"]

        # check if we have to plot the data
        hist_type = prdcfg.get("hist_type", "cumulative")
        if hist_type == "cumulative" and not data_dict["final"]:
            return None
        if hist_type == "instant" and data_dict["final"]:
            return None

        voltype_x = prdcfg["voltype_x"]
        voltype_y = prdcfg["voltype_y"]
        if voltype_x not in data_dict.keys():
            warn(
                " Field type "
                + voltype_x
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None
        if voltype_y not in data_dict.keys():
            warn(
                " Field type "
                + voltype_y
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        step_x = prdcfg.get("step_x", 0.1)
        step_y = prdcfg.get("step_y", 0.1)

        timeformat = "%Y%m%d"
        timeinfo = data_dict["timeinfo"][0]
        titl = timeinfo.strftime("%Y-%m-%d") + "\n" + voltype_x + "-" + voltype_y
        data_vals_x = data_dict[voltype_x]
        data_vals_y = data_dict[voltype_y]
        if hist_type == "instant":
            timeformat = "%Y%m%d%H%M%S"
            timeinfo = data_dict["timeinfo"][-1]
            titl = (
                timeinfo.strftime("%Y-%m-%d %H:%M:%S")
                + "\n"
                + voltype_x
                + "-"
                + voltype_y
            )
            nvols = data_dict["npoints"].size
            ind_start = np.sum(data_dict["npoints"][0 : nvols - 2])
            data_vals_x = data_vals_x[ind_start:]
            data_vals_y = data_vals_y[ind_start:]

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=timeinfo,
        )

        fname_list = make_filename(
            "2Dhistogram",
            prdcfg["dstype"],
            voltype_x + "-" + voltype_y,
            prdcfg["imgformat"],
            timeinfo=timeinfo,
            timeformat=timeformat,
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        if "_std" in voltype_x:
            bin_edges_x = np.arange(
                -1.0 - step_x / 2.0, 1.0 + step_x / 2 + step_x, step_x
            )
            values_x = data_vals_x
        else:
            field_name_x = get_fieldname_pyart(voltype_x)
            bin_edges_x, values_x = compute_histogram(
                data_vals_x, field_name_x, step=step_x
            )

        if "_std" in voltype_y:
            bin_edges_y = np.arange(
                -1.0 - step_y / 2.0, 1.0 + step_y / 2 + step_y, step_y
            )
            values_y = data_vals_y
        else:
            field_name_y = get_fieldname_pyart(voltype_y)
            bin_edges_y, values_y = compute_histogram(
                data_vals_y, field_name_y, step=step_y
            )

        hist_2d, bin_edges_x, bin_edges_y = np.histogram2d(
            values_x, values_y, bins=[bin_edges_x, bin_edges_y]
        )

        plot_scatter(
            bin_edges_x,
            bin_edges_y,
            np.ma.asarray(hist_2d),
            voltype_x,
            voltype_y,
            fname_list,
            prdcfg,
            rad1_name="",
            rad2_name="",
            titl=titl,
            cmap="viridis",
        )

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if prdcfg["type"] == "HISTOGRAM_LABELED":
        if "labeled_data_dict" not in dataset:
            return None

        data_dict = dataset["labeled_data_dict"]

        voltype = prdcfg["voltype"]
        if voltype not in data_dict.keys():
            warn(
                " Field type "
                + voltype
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None

        write_data = prdcfg.get("write_data", 1)
        step = prdcfg.get("step", 0.1)

        timeformat = "%Y%m%d"
        timeinfo = data_dict["timeinfo"][0]
        titl = timeinfo.strftime("%Y-%m-%d") + "\n" + voltype
        data_vals = data_dict[voltype]

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=timeinfo,
        )

        fname_list = make_filename(
            "histogram",
            prdcfg["dstype"],
            prdcfg["voltype"],
            prdcfg["imgformat"],
            timeinfo=timeinfo,
            timeformat=timeformat,
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        if "_std" in voltype:
            bin_edges = np.arange(-1.0 - step / 2.0, 1.0 + step / 2 + step, step)
            values = data_vals
        else:
            field_name = get_fieldname_pyart(voltype)
            bin_edges, values = compute_histogram(data_vals, field_name, step=step)
        bin_centers = bin_edges[1:] - step / 2
        hist, bin_edges = np.histogram(values, bins=bin_edges)
        plot_histogram2(
            bin_centers,
            hist,
            fname_list,
            labelx=voltype,
            labely="Number of Samples",
            titl=titl,
        )

        print("----- save to " + " ".join(fname_list))

        if write_data:
            fname = (
                savedir
                + make_filename(
                    "histogram",
                    prdcfg["dstype"],
                    prdcfg["voltype"],
                    ["csv"],
                    timeinfo=timeinfo,
                    timeformat=timeformat,
                )[0]
            )

            write_histogram(bin_edges, hist, fname, step=step)
            print("----- save to {}".format(fname))

            return fname

        return fname_list

    if prdcfg["type"] == "HISTOGRAM_CENTROIDS":
        if "labeled_data_dict" not in dataset:
            return None

        data_dict = dataset["labeled_data_dict"]

        voltype = prdcfg["voltype"]
        hydro_type = prdcfg["hydro_type"]
        if voltype not in data_dict.keys():
            warn(
                " Field type "
                + voltype
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None
        if hydro_type not in data_dict["medoids_dict"]:
            warn("No medoids where found for hydrometeor class " + hydro_type)
            return None
        ind_hydro = np.where(np.array(data_dict["hydro_names"]) == hydro_type)[0]
        ind_medoid = np.where(np.array(data_dict["var_names"]) == voltype)[0]

        write_data = prdcfg.get("write_data", 1)
        step = prdcfg.get("step", 0.1)
        dpi = 72
        if "dpi" in prdcfg["ppiImageConfig"]:
            dpi = prdcfg["ppiImageConfig"]["dpi"]

        timeformat = "%Y%m%d"
        timeinfo = data_dict["timeinfo"][0]
        titl = timeinfo.strftime("%Y-%m-%d") + "\n" + voltype + " " + hydro_type
        data_vals = data_dict[voltype]
        data_vals = data_vals[data_dict["labels"] == ind_hydro]

        medoids = np.array(data_dict["medoids_dict"][hydro_type])
        medoids = medoids[:, ind_medoid]

        if hydro_type not in data_dict["final_medoids_dict"]:
            warn("No medoid for hydrometeor class " + hydro_type)
            fmedoid = None
        else:
            fmedoid = data_dict["final_medoids_dict"][hydro_type][ind_medoid]

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=timeinfo,
        )

        fname_list = make_filename(
            "histogram",
            prdcfg["dstype"],
            hydro_type + "_" + prdcfg["voltype"],
            prdcfg["imgformat"],
            timeinfo=timeinfo,
            timeformat=timeformat,
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        if "_std" in voltype:
            bin_edges = np.arange(-1.0 - step / 2.0, 1.0 + step / 2 + step, step)
            values = data_vals
        else:
            field_name = get_fieldname_pyart(voltype)
            bin_edges, values = compute_histogram(data_vals, field_name, step=step)
        bin_centers = bin_edges[1:] - step / 2
        hist, bin_edges = np.histogram(values, bins=bin_edges)
        pos_medoids = []
        for medoid in medoids:
            ind = np.where(bin_edges <= medoid)[0][-1]
            pos_medoids.append(hist[ind])
        pos_medoids = np.array(pos_medoids)
        if fmedoid is not None:
            ind = np.where(bin_edges <= fmedoid)[0][-1]
            pos_fmedoid = hist[ind]

        (
            fig,
            ax,
        ) = plot_histogram2(
            bin_centers,
            hist,
            fname_list,
            labelx=voltype,
            labely="Number of Samples",
            titl=titl,
            save_fig=False,
        )

        ax.scatter(medoids, pos_medoids, c="g", marker="o")
        if fmedoid is not None:
            ax.plot(fmedoid, pos_fmedoid, c="r", marker="D")

        for fname in fname_list:
            fig.savefig(fname, dpi=dpi)
        plt.close(fig)

        print("----- save to " + " ".join(fname_list))

        if write_data:
            fname = (
                savedir
                + make_filename(
                    "histogram",
                    prdcfg["dstype"],
                    prdcfg["voltype"],
                    ["csv"],
                    timeinfo=timeinfo,
                    timeformat=timeformat,
                )[0]
            )

            write_histogram(bin_edges, hist, fname, step=step)
            print("----- save to {}".format(fname))

            return fname

        return fname_list

    if prdcfg["type"] == "HISTOGRAM2D_CENTROIDS":
        if "labeled_data_dict" not in dataset:
            return None

        labeled_data_dict = dataset["labeled_data_dict"]

        voltype_x = prdcfg["voltype_x"]
        voltype_y = prdcfg["voltype_y"]
        hydro_type = prdcfg["hydro_type"]
        if voltype_x not in labeled_data_dict.keys():
            warn(
                " Field type "
                + voltype_x
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None
        if voltype_y not in labeled_data_dict.keys():
            warn(
                " Field type "
                + voltype_y
                + " not available in data set. Skipping product "
                + prdcfg["type"]
            )
            return None
        if hydro_type not in labeled_data_dict["medoids_dict"]:
            warn("No medoids where found for hydrometeor class " + hydro_type)
            return None

        ind_hydro = np.where(np.array(labeled_data_dict["hydro_names"]) == hydro_type)[
            0
        ]
        ind_medoid_x = np.where(np.array(labeled_data_dict["var_names"]) == voltype_x)[
            0
        ]
        ind_medoid_y = np.where(np.array(labeled_data_dict["var_names"]) == voltype_y)[
            0
        ]

        step_x = prdcfg.get("step_x", 0.1)
        step_y = prdcfg.get("step_y", 0.1)

        timeformat = "%Y%m%d"
        timeinfo = labeled_data_dict["timeinfo"][0]
        titl = (
            timeinfo.strftime("%Y-%m-%d")
            + "\n"
            + voltype_x
            + "-"
            + voltype_y
            + " "
            + hydro_type
        )

        data_vals_x = labeled_data_dict[voltype_x]
        data_vals_y = labeled_data_dict[voltype_y]

        data_vals_x = data_vals_x[labeled_data_dict["labels"] == ind_hydro]
        data_vals_y = data_vals_y[labeled_data_dict["labels"] == ind_hydro]

        medoids = np.array(labeled_data_dict["medoids_dict"][hydro_type])
        medoids_x = medoids[:, ind_medoid_x]
        medoids_y = medoids[:, ind_medoid_y]

        if hydro_type not in labeled_data_dict["final_medoids_dict"]:
            warn("No medoid for hydrometeor class " + hydro_type)
            fmedoid_x = None
            fmedoid_y = None
        else:
            fmedoid_x = labeled_data_dict["final_medoids_dict"][hydro_type][
                ind_medoid_x
            ]
            fmedoid_y = labeled_data_dict["final_medoids_dict"][hydro_type][
                ind_medoid_y
            ]

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=timeinfo,
        )

        fname_list = make_filename(
            "2Dhistogram",
            prdcfg["dstype"],
            hydro_type + "_" + voltype_x + "-" + voltype_y,
            prdcfg["imgformat"],
            timeinfo=timeinfo,
            timeformat=timeformat,
        )

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir + fname

        if "_std" in voltype_x:
            bin_edges_x = np.arange(
                -1.0 - step_x / 2.0, 1.0 + step_x / 2 + step_x, step_x
            )
            values_x = data_vals_x
        else:
            field_name_x = get_fieldname_pyart(voltype_x)
            bin_edges_x, values_x = compute_histogram(
                data_vals_x, field_name_x, step=step_x
            )

        if "_std" in voltype_y:
            bin_edges_y = np.arange(
                -1.0 - step_y / 2.0, 1.0 + step_y / 2 + step_y, step_y
            )
            values_y = data_vals_y
        else:
            field_name_y = get_fieldname_pyart(voltype_y)
            bin_edges_y, values_y = compute_histogram(
                data_vals_y, field_name_y, step=step_y
            )

        hist_2d, bin_edges_x, bin_edges_y = np.histogram2d(
            values_x, values_y, bins=[bin_edges_x, bin_edges_y]
        )

        plot_centroids(
            bin_edges_x,
            bin_edges_y,
            np.ma.asarray(hist_2d),
            voltype_x,
            voltype_y,
            fname_list,
            prdcfg,
            titl=titl,
            medoids_x=medoids_x,
            medoids_y=medoids_y,
            fmedoid_x=fmedoid_x,
            fmedoid_y=fmedoid_y,
            cmap="viridis",
        )

        print("----- save to " + " ".join(fname_list))

        return fname_list

    if prdcfg["type"] == "WRITE_CENTROIDS":
        if "labeled_data_dict" not in dataset:
            return None

        labeled_data_dict = dataset["labeled_data_dict"]
        timeformat = "%Y%m%d"
        timeinfo = labeled_data_dict["timeinfo"][0]

        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=timeinfo,
        )

        fname = make_filename(
            "centroids",
            prdcfg["dstype"],
            "centroids",
            ["csv"],
            timeinfo=timeinfo,
            timeformat=timeformat,
        )[0]

        fname = savedir + fname

        write_centroids(
            fname,
            labeled_data_dict["final_medoids_dict"],
            labeled_data_dict["var_names"],
        )
        print("----- save to {}".format(fname))

        return fname

    if prdcfg["type"] == "SAVE_DATA":
        data_dict = dataset["data_dict"]

        # check if we have to save the data
        if not data_dict["final"]:
            return None

        timeinfo = data_dict["timeinfo"][0]
        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=timeinfo,
        )

        fname = make_filename(
            "data", prdcfg["dstype"], "centroids_data", ["npz"], timeinfo=timeinfo
        )[0]

        fname = savedir + fname

        np.savez_compressed(fname, centroids_data=data_dict)

        print("----- save to {}".format(fname))

        return fname

    if prdcfg["type"] == "SAVE_LABELED_DATA":
        if "labeled_data_dict" not in dataset:
            return None

        data_dict = dataset["labeled_data_dict"]

        timeinfo = data_dict["timeinfo"][0]
        savedir = get_save_dir(
            prdcfg["basepath"],
            prdcfg["procname"],
            dssavedir,
            prdcfg["prdname"],
            timeinfo=timeinfo,
        )

        fname = make_filename(
            "labeled_data",
            prdcfg["dstype"],
            "centroids_data",
            ["npz"],
            timeinfo=timeinfo,
        )[0]

        fname = savedir + fname

        np.savez_compressed(fname, centroids_data=data_dict)

        print("----- save to {}".format(fname))

        return fname

    warn(" Unsupported product type: " + prdcfg["type"])
    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
================================================
Pyrad: The MeteoSwiss Radar Processing framework
================================================

Welcome to Pyrad!

This program performs real time processing of the data

To run the processing framework type:
    python main_process_data_rt.py \
[config_files] --starttime [process_start_time] --endtime [process_end_time] \
--cfgpath [cfgpath] --proc_period [proc_period]

If startime or endtime are specified the program will start processing at the
specified time and end at the specified time. Otherwise the program ends when
the user interrupts it.
cfgpath is an optional argument with default: \
'$HOME/pyrad/config/processing/'
proc_period is the time that has to pass before attempting to restart the
processing in [s]
if proc_finish is not none it indicates the time the program is allowed to ran
berfore forcing it to end


Example:
    python main_process_data_rt.py 'paradiso_fvj_vol.txt' \
'paradiso_fvj_rhi.txt' --starttime '20140523000000' \
--endtime '20140523001000' --cfgpath '$HOME/pyrad/config/processing/' \
--proc_period 60 --proc_finish 120

"""

# Author: fvj
# License: BSD 3 clause

import datetime
import argparse
import atexit
import os
import sys

# Force unbuffered output
sys.stdout.reconfigure(line_buffering=True)

import traceback
from warnings import warn

from pyrad.flow.flow_control import main_cosmo_rt

print(__doc__)


def main():
    """ """

    # parse the arguments
    parser = argparse.ArgumentParser(description="Entry to Pyrad processing framework")

    # positional arguments
    parser.add_argument(
        "cfgfiles", nargs="+", type=str, help="name of main configuration file"
    )

    # keyword arguments
    parser.add_argument(
        "--starttime",
        type=str,
        default=None,
        help=(
            "starting time of the data to be processed. " + "Format "
            "YYYYMMDDhhmmss"
            ""
        ),
    )
    parser.add_argument(
        "--endtime",
        type=str,
        default=None,
        help="end time of the data to be processed. Format " "YYYYMMDDhhmmss" "",
    )
    parser.add_argument(
        "--cfgpath",
        type=str,
        default=os.path.expanduser("~") + "/pyrad/config/processing/",
        help="configuration file path",
    )

    parser.add_argument(
        "--proc_period",
        type=int,
        default=60,
        help="Period between processing rounds (s)",
    )

    parser.add_argument(
        "--proc_finish",
        type=int,
        default=None,
        help="Processing time allowed before shutdown (s)",
    )

    args = parser.parse_args()

    print(
        "====== PYRAD data processing started: %s"
        % datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
    )
    atexit.register(_print_end_msg, "====== PYRAD data processing finished: ")

    print("config path: " + args.cfgpath)
    cfgfile_list = []
    for ind, cfgfile in enumerate(args.cfgfiles):
        print("config file " + str(ind) + ": " + cfgfile)
        cfgfile_list.append(args.cfgpath + cfgfile)
    if args.starttime is not None:
        print("start time: " + args.starttime)
    else:
        print("start time not defined by user")
    if args.endtime is not None:
        print("end time: " + args.endtime)
    else:
        print("end time not defined by user")

    proc_starttime = None
    if args.starttime is not None:
        proc_starttime = datetime.datetime.strptime(args.starttime, "%Y%m%d%H%M%S")
    proc_endtime = None
    if args.endtime is not None:
        proc_endtime = datetime.datetime.strptime(args.endtime, "%Y%m%d%H%M%S")

    end_proc = False
    while not end_proc:
        try:
            end_proc = main_cosmo_rt(
                cfgfile_list,
                starttime=proc_starttime,
                endtime=proc_endtime,
                proc_period=args.proc_period,
                proc_finish=args.proc_finish,
            )
        except BaseException:
            traceback.print_exc()
            if args.proc_finish is None:
                warn("An exception occurred. " + "Restarting the real time processing")
            else:
                end_proc = True


def _print_end_msg(text):
    """
    prints end message

    Parameters
    ----------
    text : str
        the text to be printed

    Returns
    -------
    Nothing

    """
    print(
        text
        + datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
    )


# ---------------------------------------------------------
# Start main:
# ---------------------------------------------------------
if __name__ == "__main__":
    main()



"""
pyrad.io.read_data_hzt
========================

Functions for reading HZT data

.. autosummary::
    :toctree: generated/

    hzt2radar_data
    hzt2radar_coord
    get_iso0_field
    read_hzt_data
    _prepare_for_interpolation

"""

from warnings import warn
import datetime
import platform
import numpy as np
from scipy.interpolate import NearestNDInterpolator
from scipy.spatial import cKDTree

import pyart
from pyart.config import get_metadata, get_field_name

# check existence of METRANET library
try:
    METRANET_LIB = pyart.aux_io.get_library(momentms=False)
    if platform.system() == "Linux":
        METRANET_LIB = pyart.aux_io.get_library(momentms=True)
    _METRANETLIB_AVAILABLE = True
except BaseException:
    # bare exception needed to capture error
    _METRANETLIB_AVAILABLE = False

from ..io.read_data_icon import _put_radar_in_swiss_coord


def hzt2radar_data(
    radar, hzt_coord, hzt_data, slice_xy=True, field_name="height_over_iso0"
):
    """
    get the HZT value corresponding to each radar gate using nearest
    neighbour interpolation

    Parameters
    ----------
    radar : Radar
        the radar object containing the information on the position of the
        radar gates
    hzt_coord : dict
        dictionary containing the HZT coordinates
    hzt_data : dict
        dictionary containing the HZT data
    slice_xy : boolean
        if true the horizontal plane of the icon field is cut to the
        dimensions of the radar field
    field_name : str
        name of HZT fields to convert (default height_over_iso0)

    Returns
    -------
    hzt_fields : list of dict
        list of dictionary with the HZT fields and metadata

    """
    x_radar, y_radar, z_radar = _put_radar_in_swiss_coord(radar)

    x_hzt, y_hzt, ind_xmin, ind_ymin, ind_xmax, ind_ymax = _prepare_for_interpolation(
        x_radar, y_radar, hzt_coord, slice_xy=slice_xy
    )

    values = hzt_data["HZT"]["data"][
        ind_ymin : ind_ymax + 1, ind_xmin : ind_xmax + 1
    ].flatten()
    # find interpolation function
    interp_func = NearestNDInterpolator((y_hzt, x_hzt), values)

    # interpolate
    data_interp = interp_func((y_radar, x_radar))

    # put field
    field_dict = get_metadata(field_name)
    field_dict["data"] = (z_radar - data_interp).astype(float)

    return field_dict


def hzt2radar_coord(radar, hzt_coord, slice_xy=True, field_name=None):
    """
    Given the radar coordinates find the nearest HZT pixel

    Parameters
    ----------
    radar : Radar
        the radar object containing the information on the position of the
        radar gates
    hzt_coord : dict
        dictionary containing the HZT coordinates
    slice_xy : boolean
        if true the horizontal plane of the HZT field is cut to the
        dimensions of the radar field
    field_name : str
        name of the field

    Returns
    -------
    hzt_ind_field : dict
        dictionary containing a field of HZT indices and metadata

    """
    # parse the field parameters
    if field_name is None:
        field_name = get_field_name("hzt_index")

    x_radar, y_radar, _ = _put_radar_in_swiss_coord(radar)

    x_hzt, y_hzt, ind_xmin, ind_ymin, ind_xmax, _ = _prepare_for_interpolation(
        x_radar, y_radar, hzt_coord, slice_xy=slice_xy
    )

    tree = cKDTree(np.transpose((y_hzt, x_hzt)))
    _, ind_vec = tree.query(np.transpose((y_radar.flatten(), x_radar.flatten())), k=1)

    # put the index in the original icon coordinates
    nx_hzt = len(hzt_coord["x"]["data"])

    nx = ind_xmax - ind_xmin + 1

    ind_y = (ind_vec / nx).astype(int) + ind_ymin
    ind_x = (ind_vec % nx).astype(int) + ind_xmin
    ind_hzt = (ind_x + nx_hzt * ind_y).astype(int)

    hzt_ind_field = get_metadata(field_name)
    hzt_ind_field["data"] = ind_hzt.reshape(radar.nrays, radar.ngates)

    return hzt_ind_field


def get_iso0_field(hzt_data, hzt_ind, z_radar, field_name="height_over_iso0"):
    """
    Get the height over iso0 data corresponding to each radar gate
    using a precomputed look up table of the nearest neighbour

    Parameters
    ----------
    hzt_data : dict
        dictionary containing the HZT data and metadata
    hzt_ind : dict
        dictionary containing a field of HZT indices and metadata
    z_radar : ndarray
        gates altitude [m MSL]
    field_name : str
        names of HZT parameters (default height_over_iso0)

    Returns
    -------
    iso0_field : list of dict
        dictionary with the height over iso0 field and metadata

    """
    nrays, ngates = np.shape(hzt_ind["data"])
    values = hzt_data["HZT"]["data"][:, :].flatten()
    field_dict = get_metadata(field_name)
    field_dict["data"] = z_radar - values[hzt_ind["data"].flatten()].reshape(
        nrays, ngates
    ).astype(float)

    return field_dict


def read_hzt_data(fname, chy0=255.0, chx0=-160.0, read_lib="C"):
    """
    Reads iso-0 degree data from an HZT file

    Parameters
    ----------
    fname : str
        name of the file to read
    chy0, chx0: float
        south west point of grid in Swiss coordinates [km]
    read_lib : str
        Type of METRANET read library used. Can be 'C' or 'python'

    Returns
    -------
    hzt_data : dictionary
        dictionary with the data and metadata

    """
    if read_lib == "C" and _METRANETLIB_AVAILABLE:
        ret = pyart.aux_io.read_product_c(fname, physic_value=True, masked_array=True)
    elif read_lib == "python":
        ret = pyart.aux_io.read_product_py(fname, physic_value=True, masked_array=True)
    else:
        warn(
            "METRANET C-library reader not available or unknown "
            + "library type. Python library will be used"
        )
        ret = pyart.aux_io.read_product_py(fname, physic_value=True, masked_array=True)

    if ret is None:
        warn("Unable to read HZT file " + fname)
        return None

    var_data = {
        "units": "meters_above_mean_sea_level",
        "data": ret.data,
        "long_name": "height_over_iso0",
    }
    run_time = datetime.datetime.strptime(ret.header["time"], "%y%j%H%M0")
    time_data = {
        "standard_name": "time",
        "long_name": "time",
        "units": "seconds since " + run_time.strftime("%Y-%m-%d %H:%M:%S"),
        "calendar": "gregorian",
        "data": [float(ret.header["usr_forecast_hour"]) * 3600.0],
    }

    x_1 = {
        "axis": "X",
        "long_name": "x-coordinate in Swiss coordinate system",
        "standard_name": "projection_x_coordinate",
        "units": "m",
        "data": (
            (
                np.arange(int(ret.header["column"])) * float(ret.header["rect_xres"])
                + chy0
                + float(ret.header["rect_xres"]) / 2.0
            )
            * 1000.0
        ),
    }
    y_1 = {
        "axis": "Y",
        "long_name": "y-coordinate in Swiss coordinate system",
        "standard_name": "projection_y_coordinate",
        "units": "m",
        "data": (
            (
                np.arange(int(ret.header["row"])) * float(ret.header["rect_yres"])
                + chx0
                + float(ret.header["rect_yres"]) / 2.0
            )
            * 1000.0
        ),
    }

    hzt_data = {
        "HZT": var_data,
        "metadata": ret.header,
        "time": time_data,
        "x": x_1,
        "y": y_1,
    }

    return hzt_data


def _prepare_for_interpolation(x_radar, y_radar, hzt_coord, slice_xy=True):
    """
    prepares the HZT 2D volume for interpolation:
        1. if set slices the icon data to the area covered by the radar
        2. creates the x, y grid for the interpolation

    Parameters
    ----------
    x_radar, y_radar : arrays
        The Swiss coordinates of the radar
    hzt_coord : dict
        dictionary containing the HZT coordinates
    slice_xy : boolean
        if true the horizontal plane of the HZT field is cut to the
        dimensions of the radar field

    Returns
    -------
    x_hzt, y_hzt : 1D arrays
        arrays containing the flatten swiss coordinates of the HZT data in
        the area of interest [m]
    ind_xmin, ind_ymin, ind_xmax, ind_ymax : ints
        the minimum and maximum indices of each dimension

    """
    nx_hzt = len(hzt_coord["x"]["data"])
    ny_hzt = len(hzt_coord["y"]["data"])

    if slice_xy:
        # get the icon data within the radar range
        xmin = np.min(x_radar)
        xmax = np.max(x_radar)
        ymin = np.min(y_radar)
        ymax = np.max(y_radar)

        ind_xmin = np.where(hzt_coord["x"]["data"] < xmin)[0]
        if ind_xmin.size == 0:
            ind_xmin = 0
        else:
            ind_xmin = ind_xmin[-1]

        ind_xmax = np.where(hzt_coord["x"]["data"] > xmax)[0]
        if ind_xmax.size == 0:
            ind_xmax = nx_hzt - 1
        else:
            ind_xmax = ind_xmax[0]

        ind_ymin = np.where(hzt_coord["y"]["data"] < ymin)[0]
        if ind_ymin.size == 0:
            ind_ymin = 0
        else:
            ind_ymin = ind_ymin[-1]

        ind_ymax = np.where(hzt_coord["y"]["data"] > ymax)[0]
        if ind_ymax.size == 0:
            ind_ymax = ny_hzt - 1
        else:
            ind_ymax = ind_ymax[0]
    else:
        ind_xmin = 0
        ind_xmax = nx_hzt - 1
        ind_ymin = 0
        ind_ymax = ny_hzt - 1

    nx = ind_xmax - ind_xmin + 1
    ny = ind_ymax - ind_ymin + 1

    x_hzt = hzt_coord["x"]["data"][ind_xmin : ind_xmax + 1]
    y_hzt = hzt_coord["y"]["data"][ind_ymin : ind_ymax + 1]

    x_hzt = (np.broadcast_to(x_hzt.reshape(1, nx), (ny, nx))).flatten()
    y_hzt = (np.broadcast_to(y_hzt.reshape(ny, 1), (ny, nx))).flatten()

    return x_hzt, y_hzt, ind_xmin, ind_ymin, ind_xmax, ind_ymax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
================================================
main_precipitation_comparison
================================================

This program compares radar data with a point measurement sensor.

"""

# Author: fvj
# License: BSD 3 clause

import datetime
import glob
import os
import argparse

import numpy as np

from pyrad.io import read_ts_cum
from pyrad.graph import plot_scatter_comp
from pyrad.util import compute_1d_stats

print(__doc__)


def parse_args():
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        description="Compare radar data with a point measurement sensor."
    )
    parser.add_argument(
        "directories",
        help="Directories where the precipitation products are stored, multiple directories (corresponding typically to multiple radars) can be specified (comma-separated)"
        + "If multiple directories are specified, both an intercomparison between directories will be performed, as well as a "
        + "separate analysis of cumulated precipitation for every directory",
    )
    parser.add_argument(
        "--parameters", default=["RR"], nargs="+", help="List of parameters to check"
    )

    return parser.parse_args()


def find_max_min_dates(file_paths):
    dates = []

    for file_path in file_paths:
        # Extract the date component from the filename
        filename = os.path.basename(file_path)
        date_str = filename.split("_")[0]

        try:
            # Convert the date string to a datetime object
            date = datetime.datetime.strptime(date_str, "%Y%m%d")
            dates.append(date)
        except ValueError:
            # Handle files that don't have a valid date format
            print(f"Warning: '{file_path}' does not contain a valid date format.")

    if dates:
        # Find the minimum and maximum dates
        min_date = min(dates)
        max_date = max(dates)
        return min_date, max_date
    else:
        return None, None


def main():
    """Main function."""
    args = parse_args()

    fpaths = args.directories
    fpaths = fpaths.split(",")
    params = args.parameters

    np_radar_min = 6
    np_sensor_min = 6
    min_val = 0.2

    img_ext = "png"
    avg_time = 3600

    print(
        "====== precipitation comparison started: %s"
        % datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
    )

    for fpath in fpaths:
        print(f"Processing directory {fpath}")

        # Get name of pyrad output
        pyrad_name = [k for k in fpath.split("/") if len(k)][-1]

        # Find all products within file structure
        all_RR_files = glob.glob(
            os.path.join(fpath, "**", "*acc_ts_comp_POINT_MEASUREMENT_RR*.csv"),
            recursive=True,
        )

        print(f"Found {len(all_RR_files)} files with acc. precip.")

        startdate, enddate = find_max_min_dates(all_RR_files)
        print(f"Start date: {startdate}")
        print(f"End date: {enddate}")

        # Sort files by parameter
        list_params = sorted(params)[::-1]
        RR_files_by_param = {}
        for file in all_RR_files:
            for param in list_params:
                if param in file:
                    if param not in RR_files_by_param:
                        RR_files_by_param[param] = []
                    RR_files_by_param[param].append(file)

        for param in list_params:
            ts_vec = np.array([])
            val_radar = np.ma.array([])
            np_radar = np.array([])
            val_sensor = np.ma.array([])
            np_sensor = np.array([])
            for file in RR_files_by_param[param]:
                (
                    ts_aux,
                    np_radar_aux,
                    radar_value_aux,
                    np_sensor_aux,
                    sensor_value_aux,
                ) = read_ts_cum(file)
                ts_vec = np.append(ts_vec, ts_aux)
                val_radar = np.ma.append(val_radar, radar_value_aux)
                np_radar = np.append(np_radar, np_radar_aux)
                val_sensor = np.ma.append(val_sensor, sensor_value_aux)
                np_sensor = np.append(np_sensor, np_sensor_aux)
            import pdb

            pdb.set_trace()
            # filter out undesired data
            ind = np.where(
                np.logical_and(
                    np.logical_and(
                        np_radar >= np_radar_min, np_sensor >= np_sensor_min
                    ),
                    np.logical_and(val_sensor >= min_val, val_radar >= min_val),
                )
            )[0]

            val_sensor = val_sensor[ind]
            val_radar = val_radar[ind]

            # compute statistics
            stats = compute_1d_stats(val_sensor, val_radar)

            # create output image
            fpath = fpath + "RR/"
            if os.path.isdir(fpath):
                pass
            else:
                os.makedirs(fpath)
            print(f"Saving outputs to {fpath}")
            figfname = [
                startdate.strftime("%Y%m%d")
                + "-"
                + enddate.strftime("%Y%m%d")
                + "_"
                + str(avg_time)
                + "s_acc_ts_comp_"
                + param
                + "."
                + img_ext
            ]

            for i in range(len(figfname)):
                figfname[i] = fpath + figfname[i]

            labelx = "RG (mm)"
            labely = "Radar (mm)"
            titl = (
                str(avg_time)
                + " s Acc. Comp. "
                + startdate.strftime("%Y%m%d")
                + "-"
                + enddate.strftime("%Y%m%d")
            )

            metadata = (
                "npoints: "
                + str(stats["npoints"])
                + "\n"
                + "NB: "
                + "{:.2f}".format(float(stats["NB"]))
                + "\n"
                + "corr: "
                + "{:.2f}".format(float(stats["corr"]))
                + "\n"
                + "RMS: "
                + "{:.2f}".format(float(stats["RMS"]))
                + "\n"
                + "Nash: "
                + "{:.2f}".format(float(stats["Nash"]))
                + "\n"
            )

            plot_scatter_comp(
                val_sensor,
                val_radar,
                figfname,
                labelx=labelx,
                labely=labely,
                titl=titl,
                axis="equal",
                metadata=metadata,
                dpi=300,
            )


def _print_end_msg(text):
    """Prints end message."""
    print(
        text
        + datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
    )


# ---------------------------------------------------------
# Start main:
# ---------------------------------------------------------
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
================================================
main_process_data_trt
================================================

This program activates the Pyrad processing to obtain
TRT cells trajectories and obtains post-processing products

"""

# Author: fvj
# License: BSD 3 clause

import datetime
import argparse
import atexit
import os
import sys

# Force unbuffered output
sys.stdout.reconfigure(line_buffering=True)

import glob
from warnings import warn

import numpy as np

from pyrad.flow.flow_control import main as pyrad_main
from pyrad.io import get_fieldname_pyart, write_trt_cell_lightning
from pyrad.io import read_profile_ts, read_histogram_ts, read_quantiles_ts
from pyrad.io import read_trt_traj_data, read_thundertracking_info
from pyrad.graph import get_field_name, get_colobar_label
from pyrad.graph import _plot_time_range

from pyart.config import get_metadata

print(__doc__)


def main():
    """ """

    # parse the arguments
    parser = argparse.ArgumentParser(description="Entry to Pyrad processing framework")

    # positional arguments
    parser.add_argument(
        "proc_cfgfile", type=str, help="name of main configuration file"
    )

    # keyword arguments
    parser.add_argument(
        "--days",
        type=str,
        default=None,
        help="Dates to process. Format YYYY-MM-DD. Coma separated",
    )

    parser.add_argument(
        "--info_file",
        type=str,
        default="/store/msrad/radar/thundertracking/info/thundertracking_info.csv",
        help="configuration file path",
    )

    parser.add_argument(
        "--years",
        type=str,
        default=None,
        help="Years to process. If None all years in file will be processed",
    )

    parser.add_argument(
        "--max_rank", type=float, default=None, help="Max rank to process"
    )

    parser.add_argument(
        "--trtbase",
        type=str,
        default="/store/msrad/radar/trt/",
        help="name of folder containing the TRT cell data",
    )

    parser.add_argument(
        "--postproc",
        type=int,
        default=1,
        help="If true the data will be post-processed",
    )

    parser.add_argument(
        "--radarbase",
        type=str,
        default="/store/msrad/radar/pyrad_products/thundertracking/",
        help="name of folder containing the radar data",
    )

    parser.add_argument(
        "--cfgpath",
        type=str,
        default=os.path.expanduser("~") + "/pyrad/config/processing/",
        help="configuration file path",
    )

    parser.add_argument(
        "--datatypes",
        type=str,
        default="RR,hydro,KDPc,dBZc,RhoHVc,TEMP,ZDRc",
        help="Name of the polarimetric moments to process. Coma separated",
    )

    parser.add_argument(
        "--datasets",
        type=str,
        default="RR,hydro,KDPc,dBZc,RhoHVc,TEMP,ZDRc",
        help="Name of the directory containing the datasets",
    )

    parser.add_argument("--hres", type=float, default=250.0, help="Height resolution")

    parser.add_argument(
        "--path_structure",
        type=int,
        default=2,
        help="If true the data is at the cell center",
    )

    args = parser.parse_args()

    print(
        "====== PYRAD TRT data processing started: %s"
        % datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
    )
    atexit.register(_print_end_msg, "====== PYRAD TRT data processing finished: ")

    print("config path: " + args.cfgpath)
    print("config file: " + args.proc_cfgfile)
    print("trt path: " + args.trtbase)
    print("radar data path: " + args.radarbase)

    cfgfile_proc = args.cfgpath + args.proc_cfgfile
    trajtype = "trt"

    if args.days is not None:
        time_dir_list = args.days.split(",")
    else:
        # get the years to process
        years = None
        if args.years is not None:
            years = list(map(int, args.years.split(",")))

        _, max_rank, _, trt_time_start, trt_time_end = read_thundertracking_info(
            args.info_file
        )
        trt_times = np.append(trt_time_start, trt_time_end)

        trt_dates = np.array([], dtype=datetime.date)
        for trt_time in trt_times:
            trt_dates = np.append(trt_dates, trt_time.date())
        trt_dates = np.sort(np.unique(trt_dates))
        time_dir_list = []
        for rank, trt_date in zip(max_rank, trt_dates):
            if years is not None:
                if trt_date.year not in years:
                    continue

            if args.max_rank is not None:
                if rank > args.max_rank:
                    continue

            time_dir_list.append(trt_date.strftime("%Y-%m-%d"))

    if args.postproc:
        datatype_list = args.datatypes.split(",")
        dataset_list = args.datasets.split(",")

        if np.size(datatype_list) != np.size(dataset_list):
            warn(
                str(np.size(datatype_list))
                + " datatypes but "
                + str(np.size(dataset_list))
                + " dataset directories. Their number must be equal"
            )
            return

    # Find all TRT files in directory
    trt_list = []
    if args.days is not None:
        for time_dir in time_dir_list:
            trt_list.extend(
                glob.glob(args.trtbase + time_dir + "/TRTC_cell_plots/All/*.trt")
            )
            trt_list.extend(
                glob.glob(args.trtbase + time_dir + "/TRTC_cell_plots/Some/*.trt")
            )
    else:
        for time_dir in time_dir_list:
            trt_list.extend(glob.glob(args.trtbase + time_dir + "/TRTC_cell/*_tt.trt"))

    if len(trt_list) == 0:
        warn("No valid TRT files found in " + args.trtbase)
        return

    # Pyrad data processing
    trt_cell_id_list = []
    trt_file_list = []
    for fname in trt_list:
        print("processing TRT cell file " + fname)
        try:
            infostr = os.path.basename(fname).split(".")[0]
            infostr = infostr.replace("_tt", "")
            pyrad_main(cfgfile_proc, trajfile=fname, infostr=infostr, trajtype=trajtype)
            trt_cell_id_list.append(infostr)
            trt_file_list.append(fname)
        except BaseException:
            warn("Unable to process TRT cell file " + fname)

    if not args.postproc:
        return

    # plot time series and get altitude of graupel column
    if "hydro" in datatype_list:
        cell_ID_list = np.asarray([], dtype=int)
        time_list = np.asarray([], dtype=datetime.datetime)
        lon_list = np.asarray([], dtype=float)
        lat_list = np.asarray([], dtype=float)
        area_list = np.asarray([], dtype=float)
        rank_list = np.asarray([], dtype=float)
        rm_hmin_list = np.ma.asarray([], dtype=float)
        rm_hmax_list = np.ma.asarray([], dtype=float)

    for i, trt_cell_id in enumerate(trt_cell_id_list):
        print("\n\nPost-processing cell: " + trt_cell_id)
        dt_str = trt_cell_id[0:12]
        dt_cell = datetime.datetime.strptime(dt_str, "%Y%m%d%H%M")
        time_dir = dt_cell.strftime("%Y-%m-%d")
        for j, datatype in enumerate(datatype_list):
            dataset = dataset_list[j]
            if args.path_structure == 1:
                file_base2 = (
                    args.radarbase + time_dir + "/" + dataset + "_trt_center_traj/"
                )
            elif args.path_structure == 0:
                file_base2 = args.radarbase + time_dir + "/" + dataset + "_trt_traj/"
            elif args.path_structure == 2:
                file_base2 = args.radarbase + time_dir + "/trt_traj_tt/"

            field_name = get_fieldname_pyart(datatype)
            field_dict = get_metadata(field_name)
            titl = (
                "TRT cell "
                + trt_cell_id
                + "\n"
                + get_field_name(field_dict, field_name)
            )

            # plot time-height
            if args.path_structure == 2:
                flist = glob.glob(
                    file_base2
                    + "PROFILE_"
                    + dataset
                    + "/*_"
                    + trt_cell_id
                    + "_rhi_profile_*_"
                    + datatype
                    + "_hres"
                    + str(int(args.hres))
                    + ".csv"
                )
            else:
                flist = glob.glob(
                    file_base2
                    + "PROFILE/*_"
                    + trt_cell_id
                    + "_rhi_profile_*_"
                    + datatype
                    + "_hres"
                    + str(int(args.hres))
                    + ".csv"
                )

            if not flist:
                warn(
                    "No profile files found in "
                    + file_base2
                    + "PROFILE/ for TRT cell "
                    + trt_cell_id
                    + " with resolution "
                    + str(args.hres)
                )
            else:
                if args.path_structure == 1:
                    labels = ["Mean", "Min", "Max"]
                else:
                    labels = ["50.0-percentile", "25.0-percentile", "75.0-percentile"]
                    if datatype == "RhoHVc":
                        labels = [
                            "80.0-percentile",
                            "65.0-percentile",
                            "95.0-percentile",
                        ]
                    elif datatype == "hydro":
                        labels = [
                            "Mode",
                            "2nd most common",
                            "3rd most common",
                            "% points mode",
                            "% points 2nd most common",
                            "% points 3rd most common",
                        ]
                    elif datatype == "entropy" or "prop" in datatype:
                        labels = ["Mean", "Min", "Max"]

                tbin_edges, hbin_edges, _, data_ma, start_time = read_profile_ts(
                    flist, labels, hres=args.hres, t_res=None
                )

                basepath_out = os.path.dirname(flist[0])
                fname = (
                    basepath_out
                    + "/"
                    + trt_cell_id
                    + "_trt_TIME_HEIGHT_"
                    + datatype
                    + "_hres"
                    + str(args.hres)
                    + ".png"
                )

                vmin = vmax = None
                if datatype == "RhoHVc":
                    vmin = 0.95
                    vmax = 1.00

                xlabel = (
                    "time (s from " + start_time.strftime("%Y-%m-%d %H:%M:%S") + ")"
                )
                _plot_time_range(
                    tbin_edges,
                    hbin_edges,
                    data_ma,
                    field_name,
                    [fname],
                    titl=titl,
                    xlabel=xlabel,
                    ylabel="height (m MSL)",
                    figsize=[10, 8],
                    vmin=vmin,
                    vmax=vmax,
                    dpi=72,
                )

                print("----- plot to '%s'" % fname)

                # Get min and max altitude of graupel/hail area
                if datatype == "hydro":
                    (
                        traj_ID,
                        yyyymmddHHMM,
                        lon,
                        lat,
                        _,
                        _,
                        _,
                        area,
                        _,
                        _,
                        _,
                        RANKr,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                        _,
                    ) = read_trt_traj_data(trt_file_list[i])

                    hmin, hmax = get_graupel_column(
                        tbin_edges, hbin_edges, data_ma, start_time, yyyymmddHHMM
                    )

                    cell_ID_list = np.append(cell_ID_list, traj_ID)
                    time_list = np.append(time_list, yyyymmddHHMM)
                    lon_list = np.append(lon_list, lon)
                    lat_list = np.append(lat_list, lat)
                    area_list = np.append(area_list, area)
                    rank_list = np.append(rank_list, RANKr)
                    rm_hmin_list = np.ma.append(rm_hmin_list, hmin)
                    rm_hmax_list = np.ma.append(rm_hmax_list, hmax)

            # plot time-hist
            if args.path_structure == 2:
                flist = glob.glob(
                    file_base2
                    + "HISTOGRAM_"
                    + dataset
                    + "/*_"
                    + trt_cell_id
                    + "_histogram_*_"
                    + datatype
                    + ".csv"
                )
            else:
                flist = glob.glob(
                    file_base2
                    + "HISTOGRAM/*_"
                    + trt_cell_id
                    + "_histogram_*_"
                    + datatype
                    + ".csv"
                )

            if not flist:
                warn(
                    "No histogram files found in "
                    + file_base2
                    + "HISTOGRAM/ for TRT cell "
                    + trt_cell_id
                )
            else:
                tbin_edges, bin_edges, data_ma, start_time = read_histogram_ts(
                    flist, datatype, t_res=None
                )

                basepath_out = os.path.dirname(flist[0])
                fname = (
                    basepath_out
                    + "/"
                    + trt_cell_id
                    + "_trt_HISTOGRAM_"
                    + datatype
                    + ".png"
                )

                data_ma[data_ma == 0.0] = np.ma.masked
                xlabel = (
                    "time (s from " + start_time.strftime("%Y-%m-%d %H:%M:%S") + ")"
                )
                _plot_time_range(
                    tbin_edges,
                    bin_edges,
                    data_ma,
                    "frequency_of_occurrence",
                    [fname],
                    titl=titl,
                    xlabel=xlabel,
                    ylabel=get_colobar_label(field_dict, field_name),
                    vmin=0.0,
                    vmax=np.max(data_ma),
                    figsize=[10, 8],
                    dpi=72,
                )

                print("----- plot to '%s'" % fname)

            # plot quantiles
            flist = glob.glob(
                file_base2
                + "QUANTILES/*_"
                + trt_cell_id
                + "_quantiles_*_"
                + datatype
                + ".csv"
            )

            if not flist:
                warn(
                    "No quantiles files found in "
                    + file_base2
                    + "QUANTILES/ for TRT cell "
                    + trt_cell_id
                )
                continue

            tbin_edges, qbin_edges, data_ma, start_time = read_quantiles_ts(
                flist, step=5.0, qmin=0.0, qmax=100.0, t_res=None
            )

            basepath_out = os.path.dirname(flist[0])
            fname = (
                basepath_out + "/" + trt_cell_id + "_trt_QUANTILES_" + datatype + ".png"
            )

            vmin = vmax = None
            if datatype == "RhoHVc":
                vmin = 0.95
                vmax = 1.00
            xlabel = "time (s from " + start_time.strftime("%Y-%m-%d %H:%M:%S") + ")"
            _plot_time_range(
                tbin_edges,
                qbin_edges,
                data_ma,
                field_name,
                [fname],
                titl=titl,
                xlabel=xlabel,
                ylabel="Quantile",
                vmin=vmin,
                vmax=vmax,
                figsize=[10, 8],
                dpi=72,
            )

            print("----- plot to '%s'" % fname)

    if "hydro" in datatype_list:
        fname = args.trtbase + "cell_rimed_particles_column.csv"
        write_trt_cell_lightning(
            cell_ID_list,
            time_list,
            lon_list,
            lat_list,
            area_list,
            rank_list,
            rm_hmin_list,
            rm_hmax_list,
            fname,
        )

        print("----- written to '%s'" % fname)


def get_graupel_column(tbin_edges, hbin_edges, data_ma, start_time, yyyymmddHHMM):
    """
    Gets the minimum and maximum heigth of the graupel column

    Parameters
    ----------
    tbin_edges : 1D array of floats
        The time bin edges [s]
    hbin_edges : 1D array of floats
        The height bin edges [m MSL]
    data_ma : 2D array of ints
        Matrix containing the time-height hydrometeor classification
        information
    start_time : datetime object
        start time of the radar data
    yyyymmddHHMM : 1D array of datetime objects
        time steps of the TRT cell

    Returns
    -------
    hmin, hmax : 1D float arrays
        the minimum and maximum altitude of the rimed particles column

    """
    tbin_rights = tbin_edges[1:]
    hbin_lefts = hbin_edges[:-1]
    hbin_rights = hbin_edges[1:]
    H_lefts, _ = np.meshgrid(hbin_lefts, tbin_rights)
    H_lefts = np.ma.asarray(H_lefts)
    H_rights, _ = np.meshgrid(hbin_rights, tbin_rights)
    H_rights = np.ma.asarray(H_rights)
    mask = np.logical_not(np.logical_or(data_ma == 4, data_ma == 9))
    H_lefts[mask] = np.ma.masked
    H_rights[mask] = np.ma.masked
    hmin = np.ma.min(H_lefts, axis=1)
    hmax = np.ma.max(H_rights, axis=1)

    # All TRT cell time steps have a corresponding rimed column height value
    # Return the values
    if tbin_rights.size == yyyymmddHHMM.size:
        return hmin, hmax

    # Missing rimed height values. Determine those missing and put the data
    # in the right time step
    hmin_aux = np.ma.masked_all(yyyymmddHHMM.size, dtype=float)
    hmax_aux = np.ma.masked_all(yyyymmddHHMM.size, dtype=float)
    for k, dt_trt in enumerate(yyyymmddHHMM):
        for l, dt_rad in enumerate(tbin_rights):
            if dt_trt == start_time + datetime.timedelta(seconds=dt_rad):
                hmin_aux[k] = hmin[l]
                hmax_aux[k] = hmax[l]
                break
    return hmin_aux, hmax_aux


def _print_end_msg(text):
    """
    prints end message

    Parameters
    ----------
    text : str
        the text to be printed

    Returns
    -------
    Nothing

    """
    print(
        text
        + datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
    )


# ---------------------------------------------------------
# Start main:
# ---------------------------------------------------------
if __name__ == "__main__":
    main()
